/*
 * Postscript definitions.
 *
 * This file is part of abcm2ps.
 *
 * Copyright (C) 1998-2010 Jean-François Moine
 * Adapted from abc2ps, Copyright (C) 1996,1997 Michael Methfessel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <time.h>
#include <string.h>

#include "abc2ps.h"

static char ps_head[] =
	"/xymove{/x 2 index def/y 1 index def M}!\n"

	/* str showc - center at current pt */
	"/showc{dup stringwidth pop .5 mul neg 0 RM show}!\n"

	/* str showr - show right-aligned */
	"/showr{dup stringwidth pop neg 0 RM show}!\n"

	/* str showb - show in box */
	"/showb{	dup currentpoint 3 -1 roll show\n"
	"	.6 SLW\n"
	"	exch 2 sub exch 3 sub 3 -1 roll\n"
#if 1
	"	stringwidth pop 4 add\n"
	"	currentfont /ScaleMatrix get 0 get .8 mul\n"
	"	4 add rectstroke}!\n"
#else
	"	stringwidth pop 4 add fh 4 add rectstroke}!\n"
#endif

#if 0
	"/showcb{ % usage: str showcb - show centered in box\n"
	"	dup stringwidth pop dup .5 mul neg 0 RM currentpoint 4 -1 roll show\n"
	"	.6 SLW\n"
	"	exch 2 sub exch 3 sub 3 -1 roll\n"
	"	4 add fh 4 add rectstroke}!\n"
#endif

	/* x y tclef - treble clef */
	"/tclef{	M\n"
	"	-1.9 3.7 RM\n"
	"	-3.3 1.9 -3.1 6.8 2.4 8.6 RC\n"
	"	7 0 9.8 -8 4.1 -11.7 RC\n"
	"	-5.2 -2.4 -12.5 0 -13.3 6.2 RC\n"
	"	-0.7 6.4 4.15 10.5 10 15.3 RC\n"
	"	4 4 3.6 6.1 2.8 9.6 RC\n"
	"	-2.3 -1.5 -4.7 -4.8 -4.5 -8.5 RC\n"
	"	0.8 -12.2 3.4 -17.3 3.5 -26.3 RC\n"
	"	0.3 -4.4 -1.2 -6.2 -3.8 -6.2 RC\n"
	"	-3.7 -0.1 -5.8 4.3 -2.8 6.1 RC\n"
	"	3.9 1.9 6.1 -4.6 1.4 -4.8 RC\n"
	"	0.7 -1.2 4.6 -0.8 4.2 4.2 RC\n"
	"	-0.2 10.3 -3 15.7 -3.5 28.3 RC\n"
	"	0 4.1 0.6 7.4 5 10.6 RC\n"
	"	2.3 -3.2 2.9 -10 1 -12.7 RC\n"
	"	-2.4 -4.3 -11.5 -10.3 -11.8 -15 RC\n"
	"	0.4 -7 6.9 -8.5 11.7 -6.1 RC\n"
	"	3.9 3 1.3 8.8 -3.7 8.1 RC\n"
	"	-4 -0.2 -4.8 -3.1 -2.7 -5.7 RC\n"
	"	fill}!\n"

	"/stclef{gsave T .85 dup scale 0 0 tclef grestore}!\n"

	/* x y octu - upper '8' */
	"/octu{/Times-Roman 12 selectfont M -1.5 36 RM(8)show}!\n"
	/* x y octl - lower '8' */
	"/octl{/Times-Roman 12 selectfont M -3.5 -19 RM(8)show}!\n"

	/* x y bclef - bass clef */
	"/bclef{	M\n"
	"	-8.8 3.5 RM\n"
	"	6.3 1.9 10.2 5.6 10.5 10.8 RC\n"
	"	0.3 4.9 -0.5 8.1 -2.6 8.8 RC\n"
	"	-2.5 1.2 -5.8 -0.7 -5.9 -4.1 RC\n"
	"	1.8 3.1 6.1 -0.6 3.1 -3 RC\n"
	"	-3 -1.4 -5.7 2.3 -1.9 7 RC\n"
	"	2.6 2.3 11.4 0.6 10.1 -8 RC\n"
	"	-0.1 -4.6 -5 -10.2 -13.3 -11.5 RC\n"
	"	16 17.5 RM currentpoint 1.1 0 360 arc\n"
	"	-1.1 -6 RM currentpoint 1.1 0 360 arc\n"
	"	fill}!\n"

	"/sbclef{gsave T .85 dup scale 0 3 bclef grestore}!\n"

	"/cchalf{0 12 M\n"
	"	2.6 5 RL\n"
	"	2.3 -5.8 5.2 -2.4 4.7 1.6 RC\n"
	"	0.4 3.9 -3 6.7 -5.1 4 RC\n"
	"	4.1 0.5 0.9 -5.3 -0.9 -1.4 RC\n"
	"	-0.5 3.4 6.5 4.3 7.8 -0.8 RC\n"
	"	1.9 -5.6 -4.1 -9.8 -6 -5.4 RC\n"
	"	-1.6 -3 RL\n"
	"	fill}!\n"

	/* x y cclef */
	"/cclef{	gsave T\n"
	"	cchalf 0 24 T 1 -1 scale cchalf\n"
	"	-5.5 0 3 24 rectfill\n"
	"	-0.5 0 M 0 24 RL 0.7 SLW stroke grestore}!\n"

	"/scclef{gsave T .85 dup scale 0 2 cclef grestore}!\n"

	/* x y pclef */
	"/pclef{	exch 2.7 sub exch 2 add 5.4 20\n"
	"	1.4 SLW rectstroke}!\n"
	"/spclef{pclef}!\n"

	/* t dx dy x y bm - beam, depth t */
	"/bm{	M 3 copy RL neg 0 exch RL\n"
	"	neg exch neg exch RL 0 exch RL fill}!\n"

	/* str x y bnum - tuplet number / ratio */
	"/bnum{M/Times-Italic 12 selectfont showc}!\n"
	/* same with clearing below the number */
	"/bnumb{	currentgray/Times-Italic 12 selectfont\n"
	"	3 index stringwidth pop 4 add\n"
	"	dup .5 mul neg 4 index add 3 index 3 -1 roll 8\n"
	"	1.0 setgray rectfill setgray M showc}!\n"

	/* dx dy x y tubr - tuplet bracket */
	"/tubr{3 sub M 0 3 RL RL 0 -3 RL dlw stroke}!\n"
	"/tubrl{3 add M 0 -3 RL RL 0 3 RL dlw stroke}!\n"

	/* x y r00 - longa rest */
	"/r00{	xymove\n"
	"	-1.5 -6 RM currentpoint 3 12 rectfill}!\n"

	/* x y r0 - breve rest */
	"/r0{	xymove\n"
	"	-1.5 0 RM currentpoint 3 6 rectfill}!\n"

	/* x y r1 - rest */
	"/r1{	xymove\n"
	"	-3.5 3 RM currentpoint 7 3 rectfill}!\n"

	/* x y r2 - half rest */
	"/r2{	xymove\n"
	"	-3.5 0 RM currentpoint 7 3 rectfill}!\n"

	/* x y r4 - quarter rest */
	"/r4{	xymove\n"
	"	-1 8.5 RM\n"
	"	3.6 -5.1 RL\n"
	"	-2.1 -5.2 RL\n"
	"	2.2 -4.3 RL\n"
	"	-2.6 2.3 -5.1 0 -2.4 -2.6 RC\n"
	"	-4.8 3 -1.5 6.9 1.4 4.1 RC\n"
	"	-3.1 4.5 RL\n"
	"	1.9 5.1 RL\n"
	"	-1.5 3.5 RL\n"
	"	fill}!\n"

	/* 1/8 .. 1/64 rest element */
	"/r8e{	-1.5 -1.5 -2.4 -2 -3.6 -2 RC\n"
	"	2.4 2.8 -2.8 4 -2.8 1.2 RC\n"
	"	0 -2.7 4.3 -2.4 5.9 -0.6 RC\n"
	"	fill}!\n"

	/* x y r8 - eighth rest */
	"/r8{	xymove\n"
	"	.5 SLW 3.3 4 RM\n"
	"	-3.4 -9.6 RL stroke\n"
	"	x y M 3.4 4 RM r8e}!\n"

	/* x y r16 - 16th rest */
	"/r16{	xymove\n"
	"	.5 SLW 3.3 4 RM\n"
	"	-4 -15.6 RL stroke\n"
	"	x y M 3.4 4 RM r8e\n"
	"	x y M 1.9 -2 RM r8e}!\n"

	/* x y r32 - 32th rest */
	"/r32{	xymove\n"
	"	.5 SLW 4.8 10 RM\n"
	"	-5.5 -21.6 RL stroke\n"
	"	x y M 4.9 10 RM r8e\n"
	"	x y M 3.4 4 RM r8e\n"
	"	x y M 1.9 -2 RM r8e}!\n"

	/* x y r64 - 64th rest */
	"/r64{	xymove\n"
	"	.5 SLW 4.8 10 RM\n"
	"	-7 -27.6 RL stroke\n"
	"	x y M 4.9 10 RM r8e\n"
	"	x y M 3.4 4 RM r8e\n"
	"	x y M 1.9 -2 RM r8e\n"
	"	x y M 0.4 -8 RM r8e}!\n"

	/* x y r128 - 128th rest */
	"/r128{	xymove\n"
	"	.5 SLW 5.8 16 RM\n"
	"	-8.5 -33.6 RL stroke\n"
	"	x y M 5.9 16 RM r8e\n"
	"	x y M 4.4 10 RM r8e\n"
	"	x y M 2.9 4 RM r8e\n"
	"	x y M 1.4 -2 RM r8e\n"
	"	x y M -0.1 -8 RM r8e}!\n"

	/* dx dy dt - dot relative to head */
	"/dt{x y M RM currentpoint 1.2 0 360 arc fill}!\n"

	/* x y hld - fermata */
	"/hld{	1.5 add 2 copy 1.5 add M currentpoint 1.3 0 360 arc\n"
	"	M -7.5 0 RM\n"
	"	0 11.5 15 11.5 15 0 RC\n"
	"	-0.25 0 RL\n"
	"	-1.25 9 -13.25 9 -14.5 0 RC\n"
	"	fill}!\n"

	/* x y dnb - down bow */
	"/dnb{	dlw M -3.2 2 RM\n"
	"	0 7.2 RL\n"
	"	6.4 0 RM\n"
	"	0 -7.2 RL\n"
	"	currentpoint stroke M\n"
	"	-6.4 4.8 RM\n"
	"	0 2.4 RL\n"
	"	6.4 0 RL\n"
	"	0 -2.4 RL\n"
	"	fill}!\n"

	/* x y upb - up bow */
	"/upb{	dlw M -2.6 9.4 RM\n"
	"	2.6 -8.8 RL\n"
	"	2.6 8.8 RL\n"
	"	stroke}!\n"

	/* x y grm - gracing mark */
	"/grm{	M -5 2.5 RM\n"
	"	5 8.5 5.5 -4.5 10 2 RC\n"
	"	-5 -8.5 -5.5 4.5 -10 -2 RC fill}!\n"

	/* x y stc - staccato mark */
	"/stc{3 add M currentpoint 1.2 0 360 arc fill}!\n"

	/* x y emb - emphasis bar */
	"/emb{	1.2 SLW 1 setlinecap M\n"
	"	-2.5 3 RM 5 0 RL stroke 0 setlinecap}!\n"

	/* x y cpu - roll sign above head */
	"/cpu{	M -6 0 RM\n"
	"	0.4 7.3 11.3 7.3 11.7 0 RC\n"
	"	-1.3 6 -10.4 6 -11.7 0 RC fill}!\n"

	/* x y sld - slide */
	"/sld{	M -7.2 -4.8 RM\n"
	"	1.8 -0.7 4.5 0.2 7.2 4.8 RC\n"
	"	-2.1 -5 -5.4 -6.8 -7.6 -6 RC fill}!\n"

	/* x y trl - trill sign */
	"/trl{	/Times-BoldItalic 16 selectfont\n"
	"	M -4 2 RM(tr)show}!\n"

	/* x y umrd - upper mordent */
	"/umrd{	4 add M\n"
	"	2.2 2.2 RL 2.1 -2.9 RL 0.7 0.7 RL\n"
	"	-2.2 -2.2 RL -2.1 2.9 RL -0.7 -0.7 RL\n"
	"	-2.2 -2.2 RL -2.1 2.9 RL -0.7 -0.7 RL\n"
	"	2.2 2.2 RL 2.1 -2.9 RL 0.7 0.7 RL fill}!\n"

	/* x y lmrd - lower mordent */
	"/lmrd{	2 copy umrd 8 add M\n"
	"	.6 SLW 0 -8 RL stroke}!\n"

	/* str x y fng - finger (0-5) */
	"/fng{/Bookman-Demi 8 selectfont M -3 1 RM show}!\n"

	/* str x y dacs - D.C. / D.S. */
	"/dacs{/Times-Roman 16 selectfont 3 add M showc}!\n"

	/* x y brth - breath */
	"/brth{/Times-BoldItalic 30 selectfont 6 add M(,)show}!\n"

	/* str x y pf - p, f, pp, .. */
	"/pf{/Times-BoldItalic 16 selectfont 5 add M show}!\n"

	/* str x y sfz */
	"/sfz{	M -7 5 RM pop\n"
	"	/Times-Italic 14 selectfont(s)show\n"
	"	/Times-BoldItalic 16 selectfont(f)show\n"
	"	/Times-Italic 14 selectfont(z)show}!\n"

	/* x y coda - coda */
	"/coda{	1 SLW 2 add 2 copy M 0 20 RL\n"
	"	2 copy M -10 10 RM 20 0 RL stroke\n"
	"	10 add 6 0 360 arc 1.7 SLW stroke}!\n"

	/* x y sgno - segno */
	"/sgno{	M 0 3 RM currentpoint currentpoint currentpoint\n"
	"	1.5 -1.7 6.4 0.3 3 3.7 RC\n"
	"	-10.4 7.8 -8 10.6 -6.5 11.9 RC\n"
	"	4 1.9 5.9 -1.7 4.2 -2.6 RC\n"
	"	-1.3 -0.7 -2.9 1.3 -0.7 2 RC\n"
	"	-1.5 1.7 -6.4 -0.3 -3 -3.7 RC\n"
	"	10.4 -7.8 8 -10.6 6.5 -11.9 RC\n"
	"	-4 -1.9 -5.9 1.7 -4.2 2.6 RC\n"
	"	1.3 0.7 2.9 -1.3 0.7 -2 RC\n"
	"	fill\n"
	"	M 0.8 SLW -6 1.2 RM 12.6 12.6 RL stroke\n"
	"	7 add exch 6 sub exch 1.2 0 360 arc fill\n"
	"	8 add exch 6 add exch 1.2 0 360 arc fill}!\n"

	/* w x y cresc - crescendo */
	"/cresc{	1 SLW M dup 5 RM\n"
	"	defl 1 and 0 eq\n"
	"	{dup neg 4 RL 4 RL}\n"
	"	{dup neg 2.2 RL 0 3.6 RM 2.2 RL}\n"
	"	ifelse stroke}!\n"

	/* w x y dim - diminuendo */
	"/dim{	1 SLW 5 add M\n"
	"	defl 2 and 0 eq\n"
	"	{dup 4 RL neg 4 RL}\n"
	"	{dup 2.2 RL 0 3.6 RM neg 2.2 RL}\n"
	"	ifelse stroke}!\n"

	/* x y dplus - plus */
	"/dplus{	1.2 SLW 0.5 add M 0 6 RL\n"
	"	-3 -3 RM 6 0 RL stroke}!\n"

	/* x y accent - accent */
	"/accent{1.2 SLW M -4 2 RM\n"
	"	8 2 RL -8 2 RL stroke}!\n"

	/* x y turn - turn */
	"/turn{	M 5.2 8 RM\n"
	"	1.4 -0.5 0.9 -4.8 -2.2 -2.8 RC\n"
	"	-4.8 3.5 RL\n"
	"	-3 2 -5.8 -1.8 -3.6 -4.4 RC\n"
	"	1 -1.1 2 -0.8 2.1 0.1 RC\n"
	"	0.1 0.9 -0.7 1.2 -1.9 0.6 RC\n"
	"	-1.4 0.5 -0.9 4.8 2.2 2.8 RC\n"
	"	4.8 -3.5 RL\n"
	"	3 -2 5.8 1.8 3.6 4.4 RC\n"
	"	-1 1.1 -2 0.8 -2.1 -0.1 RC\n"
	"	-0.1 -0.9 0.7 -1.2 1.9 -0.6 RC\n"
	"	fill}!\n"

	/* x y trnx - turn with line through it */
	"/turnx{	2 copy turn M\n"
	"	.6 SLW 0 1.5 RM 0 9 RL stroke}!\n"

	/* x y lphr - longphrase */
	"/lphr{1.2 SLW M 0 -18 RL stroke}!\n"

	/* x y mphr - mediumphrase */
	"/mphr{1.2 SLW M 0 -12 RL stroke}!\n"

	/* x y sphr - shortphrase */
	"/sphr{1.2 SLW M 0 -6 RL stroke}!\n"

	/* w x y ltr - long trill */
	"/ltr{	gsave 4 add T\n"
	"	0 6 3 -1 roll{\n"
	/*		% first loop draws left half of squiggle; second draws right\n*/
	"		2{\n"
	"			0 0.4 M\n"
	"			2 1.9 3.4 2.3 3.9 0 curveto\n"
	"			2.1 0 lineto\n"
	"			1.9 0.8 1.4 0.7 0 -0.4 curveto\n"
	"			fill\n"
	"			180 rotate -6 0 T\n"
	"		}repeat\n"
	/*		% shift axes right one squiggle*/
	"		pop 6 0 T\n"
	"	}for\n"
	"	grestore}!\n"

	/* h x ylow arp - arpeggio */
	"/arp{gsave 90 rotate exch neg ltr grestore}!\n"

	/* x y wedge - wedge */
	"/wedge{1 add M -1.5 5 RL 3 0 RL -1.5 -5 RL fill}!\n"

	/* x y opend - 'open' sign */
	"/opend{dlw M currentpoint 3 add 2.5 -90 270 arc stroke}!\n"

	/* x y snap - 'snap' sign */
	"/snap{	dlw 2 copy M -3 6 RM\n"
	"	0 5 6 5 6 0 RC\n"
	"	0 -5 -6 -5 -6 0 RC\n"
	"	5 add M 0 -6 RL stroke}!\n"

	/* x y thumb - 'thumb' sign */
	"/thumb{	dlw 2 copy M -2.5 7 RM\n"
	"	0 6 5 6 5 0 RC\n"
	"	0 -6 -5 -6 -5 0 RC\n"
	"	2 add M 0 -4 RL stroke}!\n"

	/* n x y trem - <n> tremolo on one note */
	"/trem{	M -4.5 0 RM{\n"
	"		currentpoint\n"
	"		9 3 RL 0 -3 RL -9 -3 RL 0 3 RL\n"
	"		fill 5.4 sub M\n"
	"	}repeat}!\n"

	/* x y hl - ledger line */
	"/hl{	.8 SLW M -6 0 RM 12 0 RL stroke}!\n"
	/* x y hl1 - longer ledger line */
	"/hl1{	.8 SLW M -7 0 RM 14 0 RL stroke}!\n"
	/* x y hl2 - more longer ledger line */
	"/hl2{	.7 SLW M -9 0 RM 18 0 RL stroke}!\n"

	/* -- accidentals -- */
	/* x y sh0 - sharp sign */
	"/sh0{	gsave T .9 SLW\n"
	"	-1.2 -8.4 M 0 15.4 RL\n"
	"	1.4 -7.2 M 0 15.4 RL stroke\n"
	"	-2.6 -3 M 5.4 1.6 RL 0 -2.2 RL -5.4 -1.6 RL 0 2.2 RL fill\n"
	"	-2.6 3.4 M 5.4 1.6 RL 0 -2.2 RL -5.4 -1.6 RL 0 2.2 RL fill\n"
	"	grestore}!\n"
	/* x y ft0 - flat sign */
	"/ft0{	gsave T .8 SLW\n"
	"	-1.8 2.5 M\n"
	"	6.4 3.3 6.5 -3.6 0 -6.6 RC\n"
	"	4.6 3.9 4.5 7.6 0 5.7 RC\n"
	"	currentpoint fill M\n"
	"	0 7.1 RM 0 -12.6 RL stroke\n"
	"	grestore}!\n"
	/* x y nt0 - natural sign */
	"/nt0{	gsave T .5 SLW\n"
	"	-2 -4.3 M 0 12.2 RL\n"
	"	1.3 -7.8 M 0 12.2 RL stroke\n"
	"	2.1 SLW\n"
	"	-2 -2.9 M 3.3 0.6 RL\n"
	"	-2 2.4 M 3.3 0.6 RL stroke\n"
	"	grestore}!\n"
	/* x y dsh0 - double sharp */
	"/dsh0{	2 copy M .7 SLW\n"
	"	-2 -2 RM 4 4 RL\n"
	"	-4 0 RM 4 -4 RL stroke\n"
	"	.5 SLW 2 copy M 1.3 -1.3 RM\n"
	"	2 -0.2 RL 0.2 -2 RL -2 0.2 RL -0.2 2 RL fill\n"
	"	2 copy M 1.3 1.3 RM\n"
	"	2 0.2 RL 0.2 2 RL -2 -0.2 RL -0.2 -2 RL fill\n"
	"	2 copy M -1.3 1.3 RM\n"
	"	-2 0.2 RL -0.2 2 RL 2 -0.2 RL 0.2 -2 RL fill\n"
	"	M -1.3 -1.3 RM\n"
	"	-2 -0.2 RL -0.2 -2 RL 2 0.2 RL 0.2 2 RL fill}!\n"
	/* x y ftx - narrow flat sign */
	"/ftx{	-1.4 2.7 RM\n"
	"	5.7 3.1 5.7 -3.6 0 -6.7 RC\n"
	"	3.9 4 4 7.6 0 5.8 RC\n"
	"	currentpoint fill M\n"
	"	dlw 0 7.1 RM 0 -12.4 RL stroke}!\n"
	/* x y dft0 - double flat sign */
	"/dft0{2 copy M -2.5 0 RM ftx M 1.5 0 RM ftx}!\n"
	/* ancillary function for grace note accidentals */
	"/gsc{gsave y T .7 dup scale 0 0}!\n"

	/* some microtone accidentals */
	/* 1/4 ton sharp */
	"/sh1{	gsave T .9 SLW\n"
	"	0 -7.8 M 0 15.4 RL stroke\n"
	"	-1.8 -2.7 M 3.6 1.1 RL 0 -2.2 RL -3.6 -1.1 RL 0 2.2 RL fill\n"
	"	-1.8 3.7 M 3.6 1.1 RL 0 -2.2 RL -3.6 -1.1 RL 0 2.2 RL fill\n"
	"	grestore}!\n"
	/* 3/4 ton sharp */
	"/sh513{	gsave T .8 SLW\n"
	"	-2.5 -8.7 M 0 15.4 RL\n"
	"	0 -7.8 M 0 15.4 RL\n"
	"	2.5 -6.9 M 0 15.4 RL stroke\n"
	"	-3.7 -3.1 M 7.4 2.2 RL 0 -2.2 RL -7.4 -2.2 RL 0 2.2 RL fill\n"
	"	-3.7 3.2 M 7.4 2.2 RL 0 -2.2 RL -7.4 -2.2 RL 0 2.2 RL fill\n"
	"	grestore}!\n"
	/* 1/4 ton flat */
	"/ft1{gsave -1 1 scale exch neg exch ft0 grestore}!\n"
	/* 3/4 ton flat */
	"/ft513{2 copy gsave -1 1 scale exch neg 3 add exch M ftx grestore\n"
	"	M 1.5 0 RM ftx}!\n"

	/* str gcshow - guitar chord */
	"/gcshow{show}!\n"
	/* x y w h box - draw a box */
	"/box{.6 SLW rectstroke}!\n"
	/* str w gxshow - expand a guitar chord */
	"/find{search{pop 3 -1 roll 1 add 3 1 roll}{pop exit}ifelse}!\n"
	"/gxshow{1 index stringwidth pop\n"
	"	sub 0 2 index(	){find}loop div\n"
	"	0 9 4 -1 roll widthshow}!\n"

	/* str anshow - annotation */
	"/anshow{show}!\n"

	/* -- lyrics under notes -- */
	/* l x y wln - underscore line */
	"/wln{M .8 SLW 0 RL stroke}!\n"
	/* l x y hyph - hyphen */
	"/hyph{	.8 SLW 3 add M\n"
	"	dup cvi 40 idiv 1 add 1 index exch div\n"	/* l dx */
	"	dup 3 sub 3 1 roll\n"		/* (dx-3) l dx */
	"	dup .5 mul dup 1.5 sub 0 RM\n"	/* (dx-3) l dx (dx/2) */
	"	exch 3 -1 roll\n"		/* (dx-3) (dx/2) dx l */
	"	{pop 3 0 RL dup 0 RM}for stroke pop}!\n"
	/* str lyshow - lyrics */
	"/lyshow{show}!\n"

	/* -- bars -- */
	/* h x y bar - thin bar */
	"/bar{M dlw 0 exch RL stroke}!\n"
	/* h x y dotbar - dotted bar */
	"/dotbar{[5] 0 setdash bar [] 0 setdash}!\n"
	/* h x y thbar - thick bar */
	"/thbar{3 -1 roll 3 exch rectfill}!\n"
	/* x y rdots - repeat dots */
	"/rdots{	2 copy 9 add M currentpoint 1.2 0 360 arc\n"
	"	15 add M currentpoint 1.2 0 360 arc fill}!\n"

	/* x y csig - C timesig */
	"/csig{	M\n"
	"	1 17.3 RM\n"
	"	0.9 0 2.3 -0.7 2.4 -2.2 RC\n"
	"	-1.2 2 -3.6 -0.1 -1.6 -1.7 RC\n"
	"	2 -1 3.8 3.5 -0.8 4.7 RC\n"
	"	-2 0.4 -6.4 -1.3 -5.8 -7 RC\n"
	"	0.4 -6.4 7.9 -6.8 9.1 -0.7 RC\n"
	"	-2.3 -5.6 -6.7 -5.1 -6.8 0 RC\n"
	"	-0.5 4.4 0.7 7.5 3.5 6.9 RC\n"
	"	fill}!\n"

	/* x y ctsig - C| timesig */
	"/ctsig{dlw 2 copy csig 4 add M 0 16 RL stroke}!\n"

	/* x y xxsig - old time signatures ('o', 'o.', 'c' 'c.') */
	"/pmsig{0.3 SLW 12 add M currentpoint 5 0 360 arc stroke}!\n"
	"/pMsig{2 copy pmsig 12 add M currentpoint 1.3 0 360 arc fill}!\n"
	"/imsig{0.3 SLW 12 add 2 copy 5 add M 5 90 270 arc stroke}!\n"
	"/iMsig{2 copy imsig 12 add M currentpoint 1.3 0 360 arc fill}!\n"

	/* (top) (bot) x y tsig - time signature */
	"/tsig{	M gsave/Times-Bold 16 selectfont 1.2 1 scale\n"
	"	0 1 RM currentpoint 3 -1 roll showc\n"
	"	12 add M showc grestore}!\n"

	/* (meter) x y stsig - single time signature */
	"/stsig{	M gsave/Times-Bold 18 selectfont 1.2 1 scale\n"
	"	0 7 RM showc grestore}!\n"

	/* width n x y staff - staff with n lines*/
	"/staff{	dlw M{dup 0 RL dup neg 6 RM}repeat\n"
	"	pop stroke}!\n"

	/* l x sep0 - hline separator */
	"/sep0{dlw 0 M 0 RL stroke}!\n"

	"/hbrce{	-2.5 1 RM\n"
	"	-4.5 -4.6 -7.5 -12.2 -4.4 -26.8 RC\n"
	"	3.5 -14.3 3.2 -21.7 -2.1 -24.2 RC\n"
	"	7.4 2.4 7.3 14.2 3.5 29.5 RC\n"
	"	-2.7 9.5 -1.5 16.2 3 21.5 RC\n"
	"	fill}!\n"
	/* h x y brace */
	"/brace{	gsave T 0 0 M .01 mul 1 exch scale hbrce\n"
	"	0 -100 M 1 -1 scale hbrce grestore}!\n"

	/* h x y bracket */
	"/bracket{M dlw -5 2 RM currentpoint\n"
	"	-1.7 2 RM 10.5 -1 12 4.5 12 3.5 RC\n"
	"	0 -1 -3.5 -5.5 -8.5 -5.5 RC fill\n"
	"	3 SLW M 0 2 RM\n"
	"	0 exch neg 8 sub RL currentpoint stroke\n"
	"	dlw M -1.7 0 RM\n"
	"	10.5 1 12 -4.5 12 -3.5 RC\n"
	"	0 1 -3.5 5.5 -8.5 5.5 RC fill}!\n"

	/* nb_measures x y mrest */
	"/mrest{	gsave T 1 SLW\n"
	"	-20 6 M 0 12 RL 20 6 M 0 12 RL stroke\n"
	"	5 SLW -20 12 M 40 0 RL stroke\n"
	"	/Times-Bold 15 selectfont 0 28 M showc grestore}!\n"

	/* x y mrep - measure repeat */
	"/mrep{	2 copy 2 copy\n"
	"	M -5 15 RM currentpoint 1.4 0 360 arc\n"
	"	M 5 9 RM currentpoint 1.4 0 360 arc\n"
	"	M -7 6 RM 11 12 RL 3 0 RL -11 -12 RL -3 0 RL\n"
	"	fill}!\n"

	/* x y mrep2 - measure repeat 2 times */
	"/mrep2{	2 copy 2 copy\n"
	"	M -5 18 RM currentpoint 1.4 0 360 arc\n"
	"	M 5 6 RM currentpoint 1.4 0 360 arc fill\n"
	"	M 1.8 SLW\n"
	"	-7 4 RM 14 10 RL -14 -4 RM 14 10 RL\n"
	"	stroke}!\n"

	/* x y srep - sequence repeat */
	"/srep{	M -1 6 RM 11 12 RL 3 0 RL -11 -12 RL -3 0 RL\n"
	"	fill}!\n"

	/* str dy bracket_type dx x y repbra - repeat bracket */
	"/repbra{gsave dlw T 0 -20 M\n"
	"	0 20 3 index 1 ne{RL}{RM}ifelse 0 RL 0 ne{0 -20 RL}if stroke\n"
	"	4 exch M show grestore}!\n"

	/* pp2x pp1x p1 pp1 pp2 p2 p1 SL - slur / tie */
	"/SL{M curveto RL curveto closepath fill}!\n"

	/* pp2x pp1x p1 pp1 pp2 p2 p1 dSL - dotted slur / tie */
	"/dSL{	M [4] 0 setdash .8 SLW\n"
	"	curveto stroke [] 0 setdash\n"
	"	pop pop pop pop pop pop pop pop}!\n"

	/* -- text -- */
	"/strw{stringwidth pop w add/w exch def}!\n"
	"/jshow{w 0 32 4 -1 roll widthshow}!\n"

	/* -- note heads -- */
	/* x y hd - full head */
	"/hd{	xymove\n"
	"	3.5 2 RM\n"
	"	-2 3.5 -9 -0.5 -7 -4 RC\n"
	"	2 -3.5 9 0.5 7 4 RC fill}!\n"
	/* x y Hd - open head for half */
	"/Hd{	xymove\n"
	"	3 1.6 RM\n"
	"	-1 1.8 -7 -1.4 -6 -3.2 RC\n"
	"	1 -1.8 7 1.4 6 3.2 RC\n"
	"	0.5 0.3 RM\n"
	"	2 -3.8 -5 -7.6 -7 -3.8 RC\n"
	"	-2 3.8 5 7.6 7 3.8 RC\n"
	"	fill}!\n"
	/* x y HD - open head for whole */
	"/HD{	xymove\n"
	"	-2.7 1.4 RM\n"
	"	1.5 2.8 6.9 0 5.3 -2.7 RC\n"
	"	-1.5 -2.8 -6.9 0 -5.3 2.7 RC\n"
	"	8.3 -1.4 RM\n"
	"	0 1.5 -2.2 3 -5.6 3 RC\n"
	"	-3.4 0 -5.6 -1.5 -5.6 -3 RC\n"
	"	0 -1.5 2.2 -3 5.6 -3 RC\n"
	"	3.4 0 5.6 1.5 5.6 3 RC\n"
	"	fill}!\n"
	/* x y HDD - round breve */
	"/HDD{	dlw HD\n"
	"	x y M -6 -4 RM 0 8 RL\n"
	"	12 0 RM 0 -8 RL stroke}!\n"
	/* x y breve - square breve */
	"/breve{	xymove\n"
	"	2.5 SLW -6 -2.7 RM 12 0 RL\n"
	"	0 5.4 RM -12 0 RL stroke\n"
	"	dlw x y M -6 -5 RM 0 10 RL\n"
	"	12 0 RM 0 -10 RL stroke}!\n"
	/* x y longa */
	"/longa{	xymove\n"
	"	2.5 SLW -6 -2.7 RM 12 0 RL\n"
	"	0 5.4 RM -12 0 RL stroke\n"
	"	dlw x y M -6 -5 RM 0 10 RL\n"
	"	12 0 RM 0 -16 RL stroke}!\n"

	/* -- default percussion heads -- */
	/* x y pshhd - percussion sharp head */
	"/pshhd{/x 2 index def/y 1 index def dsh0}!\n"
	/* x y pfthd - percussion flat head */
	"/pfthd{/x 2 index def/y 1 index def dsh0\n"
	"	.7 SLW x y 4 0 360 arc stroke}!\n"
	/* same for dble sharp/flat */
	"/pdshhd{pshhd}!\n"
	"/pdfthd{pfthd}!\n"

	/* x y ghd - grace note head */
	"/ghd{	xymove\n"
	"	1.7 1 RM\n"
	"	-1 1.7 -4.5 -0.2 -3.4 -2 RC\n"
	"	1 -1.7 4.5 0.2 3.4 2 RC fill}!\n"

	/* dx dy gua / gda - acciaccatura */
	"/gua{x y M -1 4 RM RL stroke}!\n"
	"/gda{x y M -5 -4 RM RL stroke}!\n"

	/* x y ghl - grace note ledger line */
	"/ghl{	.6 SLW M -3 0 RM 6 0 RL stroke}!\n"

	/* x1 y2 x2 y2 x3 y3 x0 y0 gsl - grace note slur */
	"/gsl{dlw M curveto stroke}!\n"

	/* UTF-8 translations */
	"/utf-array[\n"
	"<20>	/space\n"
	"<21>	/exclam\n"
	"<22>	/quotedbl\n"
	"<23>	/numbersign\n"
	"<24>	/dollar\n"
	"<25>	/percent\n"
	"<26>	/ampersand\n"
	"<27>	/quoteright\n"
	"<28>	/parenleft\n"
	"<29>	/parenright\n"
	"<2a>	/asterisk\n"
	"<2b>	/plus\n"
	"<2c>	/comma\n"
	"<2d>	/hyphen\n"
	"<2e>	/period\n"
	"<2f>	/slash\n"
	"<30>	/zero\n"
	"<31>	/one\n"
	"<32>	/two\n"
	"<33>	/three\n"
	"<34>	/four\n"
	"<35>	/five\n"
	"<36>	/six\n"
	"<37>	/seven\n"
	"<38>	/eight\n"
	"<39>	/nine\n"
	"<3a>	/colon\n"
	"<3b>	/semicolon\n"
	"<3c>	/less\n"
	"<3d>	/equal\n"
	"<3e>	/greater\n"
	"<3f>	/question\n"
	"<40>	/at\n"
	"<41>	/A\n"
	"<42>	/B\n"
	"<43>	/C\n"
	"<44>	/D\n"
	"<45>	/E\n"
	"<46>	/F\n"
	"<47>	/G\n"
	"<48>	/H\n"
	"<49>	/I\n"
	"<4a>	/J\n"
	"<4b>	/K\n"
	"<4c>	/L\n"
	"<4d>	/M\n"
	"<4e>	/N\n"
	"<4f>	/O\n"
	"<50>	/P\n"
	"<51>	/Q\n"
	"<52>	/R\n"
	"<53>	/S\n"
	"<54>	/T\n"
	"<55>	/U\n"
	"<56>	/V\n"
	"<57>	/W\n"
	"<58>	/X\n"
	"<59>	/Y\n"
	"<5a>	/Z\n"
	"<5b>	/bracketleft\n"
	"<5c>	/backslash\n"
	"<5d>	/bracketright\n"
	"<5e>	/asciicircum\n"
	"<5f>	/underscore\n"
	"<60>	/quoteleft\n"
	"<61>	/a\n"
	"<62>	/b\n"
	"<63>	/c\n"
	"<64>	/d\n"
	"<65>	/e\n"
	"<66>	/f\n"
	"<67>	/g\n"
	"<68>	/h\n"
	"<69>	/i\n"
	"<6a>	/j\n"
	"<6b>	/k\n"
	"<6c>	/l\n"
	"<6d>	/m\n"
	"<6e>	/n\n"
	"<6f>	/o\n"
	"<70>	/p\n"
	"<71>	/q\n"
	"<72>	/r\n"
	"<73>	/s\n"
	"<74>	/t\n"
	"<75>	/u\n"
	"<76>	/v\n"
	"<77>	/w\n"
	"<78>	/x\n"
	"<79>	/y\n"
	"<7a>	/z\n"
	"<7b>	/braceleft\n"
	"<7c>	/bar\n"
	"<7d>	/braceright\n"
	"<7e>	/asciitilde\n"
	"<c2a1>	/exclamdown\n"
	"<c2a2>	/cent\n"
	"<c2a3>	/sterling\n"
	"<c2a4>	/currency\n"
	"<c2a5>	/yen\n"
	"<c2a6>	/brokenbar\n"
	"<c2a7>	/section\n"
	"<c2a8>	/dieresis\n"
	"<c2a9>	/copyright\n"
	"<c2aa>	/ordfeminine\n"
	"<c2ab>	/guillemotleft\n"
	"<c2ac>	/logicalnot\n"
	"<c2ae>	/registered\n"
	"<c2b0>	/degree\n"
	"<c2b1>	/plusminus\n"
	"<c2b2>	/twosuperior\n"
	"<c2b3>	/threesuperior\n"
	"<c2b4>	/acute\n"
	"<c2b6>	/paragraph\n"
	"<c2b8>	/cedilla\n"
	"<c2b9>	/onesuperior\n"
	"<c2ba>	/ordmasculine\n"
	"<c2bb>	/guillemotright\n"
	"<c2bc>	/onequarter\n"
	"<c2bd>	/onehalf\n"
	"<c2be>	/threequarters\n"
	"<c2bf>	/questiondown\n"
	"<c380>	/Agrave\n"
	"<c381>	/Aacute\n"
	"<c382>	/Acircumflex\n"
	"<c383>	/Atilde\n"
	"<c384>	/Adieresis\n"
	"<c385>	/Aring\n"
	"<c386>	/AE\n"
	"<c387>	/Ccedilla\n"
	"<c388>	/Egrave\n"
	"<c389>	/Eacute\n"
	"<c38a>	/Ecircumflex\n"
	"<c38b>	/Edieresis\n"
	"<c38c>	/Igrave\n"
	"<c38d>	/Iacute\n"
	"<c38e>	/Icircumflex\n"
	"<c38f>	/Idieresis\n"
	"<c390>	/Eth\n"
	"<c391>	/Ntilde\n"
	"<c392>	/Ograve\n"
	"<c393>	/Oacute\n"
	"<c394>	/Ocircumflex\n"
	"<c395>	/Otilde\n"
	"<c396>	/Odieresis\n"
	"<c397>	/multiply\n"
	"<c398>	/Oslash\n"
	"<c399>	/Ugrave\n"
	"<c39a>	/Uacute\n"
	"<c39b>	/Ucircumflex\n"
	"<c39c>	/Udieresis\n"
	"<c39d>	/Yacute\n"
	"<c39e>	/Thorn\n"
	"<c39f>	/germandbls\n"
	"<c3a0>	/agrave\n"
	"<c3a1>	/aacute\n"
	"<c3a2>	/acircumflex\n"
	"<c3a3>	/atilde\n"
	"<c3a4>	/adieresis\n"
	"<c3a5>	/aring\n"
	"<c3a6>	/ae\n"
	"<c3a7>	/ccedilla\n"
	"<c3a8>	/egrave\n"
	"<c3a9>	/eacute\n"
	"<c3aa>	/ecircumflex\n"
	"<c3ab>	/edieresis\n"
	"<c3ac>	/igrave\n"
	"<c3ad>	/iacute\n"
	"<c3ae>	/icircumflex\n"
	"<c3af>	/idieresis\n"
	"<c3b0>	/eth\n"
	"<c3b1>	/ntilde\n"
	"<c3b2>	/ograve\n"
	"<c3b3>	/oacute\n"
	"<c3b4>	/ocircumflex\n"
	"<c3b5>	/otilde\n"
	"<c3b6>	/odieresis\n"
	"<c3b7>	/divide\n"
	"<c3b8>	/oslash\n"
	"<c3b9>	/ugrave\n"
	"<c3ba>	/uacute\n"
	"<c3bb>	/ucircumflex\n"
	"<c3bc>	/udieresis\n"
	"<c3bd>	/yacute\n"
	"<c3be>	/thorn\n"
	"<c3bf>	/ydieresis\n"
	"<c480>	/Amacron\n"
	"<c481>	/amacron\n"
	"<c482>	/Abreve\n"
	"<c483>	/abreve\n"
	"<c484>	/Aogonek\n"
	"<c485>	/aogonek\n"
	"<c486>	/Cacute\n"
	"<c487>	/cacute\n"
	"<c488>	/Ccircumflex\n"
	"<c489>	/ccircumflex\n"
	"<c48a>	/Cdotaccent\n"
	"<c48b>	/cdotaccent\n"
	"<c48c>	/Ccaron\n"
	"<c48d>	/ccaron\n"
	"<c48e>	/Dcaron\n"
	"<c48f>	/dcaron\n"
	"<c490>	/Dcroat\n"
	"<c491>	/dcroat\n"
	"<c492>	/Emacron\n"
	"<c493>	/emacron\n"
	"<c494>	/Ebreve\n"
	"<c495>	/ebreve\n"
	"<c496>	/Edotaccent\n"
	"<c497>	/edotaccent\n"
	"<c498>	/Eogonek\n"
	"<c499>	/eogonek\n"
	"<c49a>	/Ecaron\n"
	"<c49b>	/ecaron\n"
	"<c49c>	/Gcircumflex\n"
	"<c49d>	/gcircumflex\n"
	"<c49e>	/Gbreve\n"
	"<c49f>	/gbreve\n"
	"<c4a0>	/Gdotaccent\n"
	"<c4a1>	/gdotaccent\n"
	"<c4a2>	/Gcommaaccent\n"
	"<c4a3>	/gcommaaccent\n"
	"<c4a4>	/Hcircumflex\n"
	"<c4a5>	/hcircumflex\n"
	"<c4a6>	/Hbar\n"
	"<c4a7>	/hbar\n"
	"<c4a8>	/Itilde\n"
	"<c4a9>	/itilde\n"
	"<c4aa>	/Imacron\n"
	"<c4ab>	/imacron\n"
	"<c4ac>	/Ibreve\n"
	"<c4ad>	/ibreve\n"
	"<c4ae>	/Iogonek\n"
	"<c4af>	/iogonek\n"
	"<c4b0>	/Idotaccent\n"
	"<c4b1>	/dotlessi\n"
	"<c4b2>	/IJ\n"
	"<c4b3>	/ij\n"
	"<c4b4>	/Jcircumflex\n"
	"<c4b5>	/jcircumflex\n"
	"<c4b6>	/Kcommaaccent\n"
	"<c4b7>	/kcommaaccent\n"
	"<c4b8>	/kgreenlandic\n"
	"<c4b9>	/Lacute\n"
	"<c4ba>	/lacute\n"
	"<c4bb>	/Lcommaaccent\n"
	"<c4bc>	/lcommaaccent\n"
	"<c4bd>	/Lcaron\n"
	"<c4be>	/lcaron\n"
	"<c4bf>	/Ldot\n"
	"<c581>	/Lslash\n"
	"<c582>	/lslash\n"
	"<c583>	/Nacute\n"
	"<c584>	/nacute\n"
	"<c585>	/Ncedilla\n"
	"<c586>	/ncedilla\n"
	"<c587>	/Ncaron\n"
	"<c588>	/ncaron\n"
	"<c58a>	/Eng\n"
	"<c58b>	/eng\n"
	"<c58c>	/Omacron\n"
	"<c58d>	/omacron\n"
	"<c58e>	/Obreve\n"
	"<c58f>	/obreve\n"
	"<c590>	/Ohungarumlaut\n"
	"<c591>	/ohungarumlaut\n"
	"<c594>	/Racute\n"
	"<c595>	/racute\n"
	"<c596>	/Rcedilla\n"
	"<c597>	/rcedilla\n"
	"<c598>	/Rcaron\n"
	"<c599>	/rcaron\n"
	"<c59a>	/Sacute\n"
	"<c59b>	/sacute\n"
	"<c59c>	/Scircumflex\n"
	"<c59d>	/scircumflex\n"
	"<c59e>	/Scedilla\n"
	"<c59f>	/scedilla\n"
	"<c5a0>	/Scaron\n"
	"<c5a1>	/scaron\n"
	"<c5a2>	/Tcedilla\n"
	"<c5a3>	/tcedilla\n"
	"<c5a4>	/Tcaron\n"
	"<c5a5>	/tcaron\n"
	"<c5a6>	/Tbar\n"
	"<c5a7>	/tbar\n"
	"<c5a8>	/Utilde\n"
	"<c5a9>	/utilde\n"
	"<c5aa>	/Umacron\n"
	"<c5ab>	/umacron\n"
	"<c5ac>	/Ubreve\n"
	"<c5ad>	/ubreve\n"
	"<c5ae>	/Uring\n"
	"<c5af>	/uring\n"
	"<c5b0>	/Uhungarumlaut\n"
	"<c5b1>	/uhungarumlaut\n"
	"<c5b2>	/Uogonek\n"
	"<c5b3>	/uogonek\n"
	"<c5b9>	/Zacute\n"
	"<c5ba>	/zacute\n"
	"<c5bb>	/Zdotaccent\n"
	"<c5bc>	/zdotaccent\n"
	"<c5bd>	/Zcaron\n"
	"<c5be>	/zcaron\n"
	"<cb87>	/caron\n"
	"<cb98>	/breve\n"
	"<cb99>	/dotaccent\n"
	"<cb9b>	/ogonek\n"
	"<cb9d>	/hungarumlaut\n"
	"<e280a6> /ellipsis\n"
	"<e282ac> /Euro\n"
	"]def\n"

	/* extra characters and drawing functions */
	"/extra-array[\n"
	"<e299ad> /flat\n"	/* or /uni266D */
	"<e299ae> /natural\n"	/* or /uni266E */
	"<e299af> /sharp\n"	/* or /uni266F */
	"<f09d84aa> /dsharp\n"
	"<f09d84ab> /dflat\n"
	"]def\n"

	"/extra-draw<<\n"
	"	/.notdef{ }\n"
	"	/sharp{56 dup scale 3 6 sh0}\n"
	"	/flat{56 dup scale 3 5 ft0}\n"
	"	/natural{56 dup scale 3 5 nt0}\n"
	"	/dsharp{56 dup scale 3 5 dsh0}\n"
	"	/dflat{56 dup scale 3 5 dft0}\n"
	">>def\n"

	/* x y showerror */
	"/showerror{	gsave 1 0.7 0.7 setrgbcolor 2.5 SLW 2 copy newpath\n"
	"	30 0 360 arc stroke grestore}!\n"

	"0 setlinecap 0 setlinejoin\n";

/* -- output the common CMAP -- */
/* This output must occurs after user PostScript definitions because
 * these ones may change the default behaviour */
void define_cmap(void)
{
	static char cmap[] =

	/* font for extra characters */
	"/Extra 10 dict begin\n"
	"	/FontType 3 def\n"
	"	/FontMatrix[.001 0 0 .001 0 0]def\n"
	"	/Encoding 256 array def\n"
	"	/FontBBox[0 0 1000 1000]def\n"
	"	/BuildGlyph{\n"
	"		350 0 0 0 500 700 setcachedevice\n"
	"		userdict/extra-draw get exch\n"
	"		2 copy known not{pop/.notdef}if\n"
	"		get exec pop\n"
	"	}!\n"
	"	currentdict\n"
	"end\n"
	"definefont pop\n"

	/* CMAP definition */
	"/CIDInit/ProcSet findresource begin\n"
	"    12 dict begin\n"
	"	begincmap\n"
	"		/CMapName/CMAP-UTF8 def\n"
	"		/CMapType 1 def\n"
	"		/CIDSystemInfo[null]def\n"
	"		/WMode 0 def\n"
					/* main font */
	"		3 begincodespacerange\n"
	"			<20>	<7f>\n"
	"			<c280>	<dfbf>\n"
	"			<e28080> <e28fbf>\n"
	"		endcodespacerange\n"
	"		utf-array length 2 idiv\n"	/* not used by ghostscript */
	"		beginbfchar\n"
	"			utf-array aload pop\n"
	"		endbfchar\n"
					/* extra font */
	"		1 usefont\n"
	"		2 begincodespacerange\n"
	"			<e299ad> <e299af>\n"
	"			<f09d84aa> <f09d84ab>\n"
	"		endcodespacerange\n"
	"		extra-array length 2 idiv\n"
	"		beginbfchar\n"
	"			extra-array aload pop\n"
	"		endbfchar\n"
	"	endcmap\n"
	"    CMapName currentdict/CMap defineresource pop\n"
	"    end\n"
	"end\n";

	fputs(cmap, fout);
}

/* -- define a font -- */
void define_font(char name[],
		 int num,
		 int enc)
{
	if (enc == 0)		/* utf-8 */
		fprintf(fout, "/%s-utf8/CMAP-UTF8[/%s /Extra]composefont pop\n"
#if 1
			"/F%d{/%s-utf8 exch selectfont}!\n",
#else
			"/F%d{dup .8 mul/fh exch def/%s-utf8 exch selectfont}!\n",
#endif
			name, name, num, name);
	else			/* native encoding */
#if 1
		fprintf(fout, "/F%d{/%s exch selectfont}!\n",
#else
		fprintf(fout, "/F%d{dup .8 mul/fh exch def/%s exch selectfont}!\n",
#endif
			num, name);
}

/* -- output the symbol definitions -- */
void define_symbols(void)
{
	fputs(ps_head, fout);

	/* len su - up stem */
	fprintf(fout, "/su{dlw x y M %.1f %.1f RM %.1f sub 0 exch RL stroke}!\n",
		STEM_XOFF, STEM_YOFF, STEM_YOFF);

	/* len sd - down stem */
	fprintf(fout, "/sd{dlw x y M %.1f %.1f RM %.1f add 0 exch RL stroke}!\n",
		-STEM_XOFF, -STEM_YOFF, STEM_YOFF);

	/* n len sfu - stem and n flags up */
	fprintf(fout, "/sfu{	dlw x y M %.1f %.1f RM\n"
		"	%.1f sub 0 exch RL currentpoint stroke\n"
		"	M dup 1 eq{\n"
		"		pop\n"
		"		0.6 -5.6 9.6 -9 5.6 -18.4 RC\n"
		"		1.6 6 -1.3 11.6 -5.6 12.8 RC\n"
		"		fill\n"
		"	  }{\n"
		"		1 sub{	currentpoint\n"
		"			0.9 -3.7 9.1 -6.4 6 -12.4 RC\n"
		"			1 5.4 -4.2 8.4 -6 8.4 RC\n"
		"			fill 5.4 sub M\n"
		"		}repeat\n"
		"		1.2 -3.2 9.6 -5.7 5.6 -14.6 RC\n"
		"		1.6 5.4 -1 10.2 -5.6 11.4 RC\n"
		"		fill\n"
		"	  }ifelse}!\n",
		STEM_XOFF, STEM_YOFF, STEM_YOFF);

	/* n len sfd - stem and n flags down */
	fprintf(fout, "/sfd{	dlw x y M %.1f %.1f RM\n"
		"	%.1f add 0 exch RL currentpoint stroke\n"
		"	M dup 1 eq{\n"
		"		pop\n"
		"		0.6 5.6 9.6 9 5.6 18.4 RC\n"
		"		1.6 -6 -1.3 -11.6 -5.6 -12.8 RC\n"
		"		fill\n"
		"	  }{\n"
		"		1 sub{	currentpoint\n"
		"			0.9 3.7 9.1 6.4 6 12.4 RC\n"
		"			1 -5.4 -4.2 -8.4 -6 -8.4 RC\n"
		"			fill 5.4 add M\n"
		"		}repeat\n"
		"		1.2 3.2 9.6 5.7 5.6 14.6 RC\n"
		"		1.6 -5.4 -1 -10.2 -5.6 -11.4 RC\n"
		"		fill\n"
		"	  }ifelse}!\n",
		-STEM_XOFF, -STEM_YOFF, STEM_YOFF);

	/* n len sfs - stem and n straight flag down */
	fprintf(fout, "/sfs{	dup 0 lt{\n"
		"		dlw x y M -%.1f -%.1f RM\n"
		"		%.1f add 0 exch RL currentpoint stroke\n"
		"		M{	currentpoint\n"
		"			7 %.1f RL\n"
		"			0 %.1f RL\n"
		"			-7 -%.1f RL\n"
		"			fill 5.4 add M\n"
		"		}repeat\n"
		"	}{\n"
		"		dlw x y M %.1f %.1f RM\n"
		"		%.1f sub 0 exch RL currentpoint stroke\n"
		"		M{	currentpoint\n"
		"			7 -%.1f RL\n"
		"			0 -%.1f RL\n"
		"			-7 %.1f RL\n"
		"			fill 5.4 sub M\n"
		"		}repeat\n"
		"	}ifelse}!\n",
		STEM_XOFF, STEM_YOFF, STEM_YOFF,
		BEAM_DEPTH, BEAM_DEPTH, BEAM_DEPTH,
		STEM_XOFF, STEM_YOFF, STEM_YOFF,
		BEAM_DEPTH, BEAM_DEPTH, BEAM_DEPTH);

	/* len gu - grace note stem up */
	fprintf(fout, "/gu{	.6 SLW x y M\n"
		"	%.1f 0 RM 0 exch RL stroke}!\n"

	/* len gd - grace note stem down */
		"/gd{	.6 SLW x y M\n"
		"	%.1f 0 RM 0 exch RL stroke}!\n",
		GSTEM_XOFF, -GSTEM_XOFF);

	/* n len sgu - gnote stem and n flag up */
	fprintf(fout, "/sgu{	.6 SLW x y M %.1f 0 RM\n"
		"	0 exch RL currentpoint stroke\n"
		"	M dup 1 eq{\n"
		"		pop\n"
		"		0.6 -3.4 5.6 -3.8 3 -10 RC\n"
		"		1.2 4.4 -1.4 7 -3 7 RC\n"
		"		fill\n"
		"	  }{\n"
		"		{	currentpoint\n"
		"			1 -3.2 5.6 -2.8 3.2 -8 RC\n"
		"			1.4 4.8 -2.4 5.4 -3.2 5.2 RC\n"
		"			fill 3.5 sub M\n"
		"		}repeat\n"
		"	  }ifelse}!\n",
		GSTEM_XOFF);

	/* n len sgd - gnote stem and n flag down */
	fprintf(fout, "/sgd{	.6 SLW x y M %.1f 0 RM\n"
		"	0 exch RL currentpoint stroke\n"
		"	M dup 1 eq{\n"
		"		pop\n"
		"		0.6 3.4 5.6 3.8 3 10 RC\n"
		"		1.2 -4.4 -1.4 -7 -3 -7 RC\n"
		"		fill\n"
		"	  }{\n"
		"		{	currentpoint\n"
		"			1 3.2 5.6 2.8 3.2 8 RC\n"
		"			1.4 -4.8 -2.4 -5.4 -3.2 -5.2 RC\n"
		"			fill 3.5 add M\n"
		"		}repeat\n"
		"	  }ifelse}!\n",
		-GSTEM_XOFF);

	/* n len sgs - gnote stem and n straight flag up */
	fprintf(fout, "/sgs{	.6 SLW x y M %.1f 0 RM\n"
		"	0 exch RL currentpoint stroke\n"
		"	M{	currentpoint\n"
		"		3 -1.5 RL 0 -2 RL -3 1.5 RL\n"
		"		closepath fill 3 sub M\n"
		"	}repeat}!\n",
		GSTEM_XOFF);
}
